/*
 * Copyright (c) 2018 CompuLab Ltd. https://www.compulab.com/
 *
 * Licensed under 3-Clause BSD License, see file LICENSE in this source tree.
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpsse.h>
#include <getopt.h>
#include <inttypes.h>
#include <errno.h>
#include <time.h>
#include <pthread.h>
#include "mcp251x.h"

#define CAN_ID_SIZE 4 //number of registers to store SID + EID
#define CAN_FILTER_NUM 5 //There is 5 receive filter on the chip
#define CAN_RXBUF_NUM 2 //There is 5 receive filter on the chip
#define CAN_TXBUF_NUM 3 //There is 5 receive filter on the chip

#define VER "0.0-alpha"

/*
* Send a single cmd sequence to mcp251x chip over SPI.
* Returns MPSSE_OK on success.
* Returns MPSSE_FAIL on failure.
*/
static int single_cmd(struct mpsse_context *slave, char *buf, int len)
{
	int retval;

	Start(slave);
	retval = Write(slave, buf, len);
	Stop(slave);
	return retval;
}
static int rts_cmd(struct mpsse_context *slave, int bufnum)
{
	char buf[1];//Buffer to send

	buf[0] = INSTRUCTION_RTS(bufnum);
	return single_cmd(slave, buf, sizeof(buf));
}
static int reset_cmd(struct mpsse_context *slave)
{
	char buf[1];//Buffer to send

	buf[0] = INSTRUCTION_RESET;
	return single_cmd(slave, buf, sizeof(buf));
}
/*
* Modifies given as mask/data bits in a given reg of mcp251x.
* Returns MPSSE_OK on success.
* Returns MPSSE_FAIL on failure.
*/
static int bit_mod_reg(struct mpsse_context *slave, uint8_t reg, uint8_t mask,
		uint8_t val)
{
	char buf[4];//Buffer to send

	buf[0] = INSTRUCTION_BIT_MODIFY;
	buf[1] = reg;
	buf[2] = mask;
	buf[3] = val;
	return single_cmd(slave, buf, sizeof(buf));
}
/*
* Writes a given value to a given register (8bit) of mcp251x chip over SPI.
* Returns MPSSE_OK on success.
* Returns MPSSE_FAIL on failure.
*/
static int write_reg(struct mpsse_context *slave, uint8_t reg, uint8_t val)
{
	char buf[3];//Buffer to send

	buf[0] = INSTRUCTION_WRITE;
	buf[1] = reg;
	buf[2] = val;
	return single_cmd(slave, buf, sizeof(buf));
}
/*
* Send a given frame (size must be 13 bytes) over a given transmit buffer.
* of a given interface (bufnum must be 0,2,3)
* Returns MPSSE_OK on success.
* Returns MPSSE_FAIL on failure.
*/
#define CAN_MAX_FRAME 13
static int send_frame(struct mpsse_context *slave, int bufnum, char *frame)
{
	int retval;
	char buf[CAN_MAX_FRAME+1];//Buffer to send

	buf[0] = INSTRUCTION_LOAD_TXB(bufnum);
	memcpy(&buf[1], frame, CAN_MAX_FRAME);
	retval = single_cmd(slave, buf, sizeof(buf));
	if(retval)
		return retval;
	return bit_mod_reg(slave, TXBCTRL(bufnum),
			TXBCTRL_TXREQ, TXBCTRL_TXREQ);
}

/*
* Reads a given register (8bit) of mcp251x chip over SPI.
* Returns a positive value 0x000x of the register or a negative errorcode.
*/
static int read_reg(struct mpsse_context *slave, uint8_t reg)
{
	char *data = NULL;// Receive buffer
	char buf[3];//Buffer to send
	int retval;

	buf[0] = INSTRUCTION_READ;
	buf[1] = reg;
	buf[2] = 0;

	//The read value will be found in the [2] entry of the buffer
	Start(slave);
	data = Transfer(slave, buf, sizeof(buf));
	Stop(slave);

	if(!data){
		retval = -MPSSE_FAIL;
	}
	else{
		retval = data[2] & 0xff;
		free(data);
	}
	return retval;
}
/*
* Reads RX status of mcp251x chip over SPI.
* Returns a positive value 0x000x of the register or a negative errorcode.
*/
static int read_rx_status(struct mpsse_context *slave)
{
	char *data = NULL;// Receive buffer
	char buf[3];//Buffer to send
	int retval;

	buf[0] = INSTRUCTION_RX_STATUS;
	buf[1] = 0;

	//The read value will be found in the [2] entry of the buffer
	Start(slave);
	data = Transfer(slave, buf, sizeof(buf));
	Stop(slave);

	if(!data){
		retval = -MPSSE_FAIL;
	}
	else{
		retval = data[1] & 0xff;
		free(data);
	}
	return retval;
}/*
* Reads a given RX buffer (0 or 1) of mcp251x chip over SPI.
* Returns a pointer to the read data on success or NULL on failure.
*/
static char *read_rx_buffer(struct mpsse_context *slave, int bufnum)
{
	char *data = NULL;// Receive buffer
	char buf[CAN_MAX_FRAME+1];//Buffer to send/size to read
	int retval;
	if(0 != bufnum && 1 != bufnum)
		return NULL;
	buf[0] = INSTRUCTION_READ_RXB(bufnum);

	//The read value will be found in the [2] entry of the buffer
	Start(slave);
	retval = Write(slave, buf, 1);
	if(!retval)
		data = Read(slave, CAN_MAX_FRAME );
	Stop(slave);
	return data;
}
/*
* Reads a given count of contiguous registers of mcp251x chip over SPI.
* Returns a pointer to the read data on success or NULL on failure.
*/
static char* read_reg_seq(struct mpsse_context *slave, // mcp251x chip
		uint8_t reg, //First register
		int count) //How many regs to read
{
	char *data = NULL;// Receive buffer
	char outbuf[128];//Buffer to send
	int retval;

	outbuf[0] = INSTRUCTION_READ;
	outbuf[1] = reg;
	Start(slave);
	data = Transfer(slave, outbuf, count + 2);
	Stop(slave);
	return data;
}

/*
 * CAN bit-timing parameters
 *
 * For further information, please read chapter 8 "BIT TIMING
 * REQUIREMENTS" of the "Bosch CAN Specification version 2.0"
 * at http://www.semiconductors.bosch.de/pdf/can2spec.pdf.
 */
struct can_bittiming {
	uint32_t bitrate;		/* Bit-rate in bits/second */
	uint32_t sample_point;	/* Sample point in one-tenth of a percent */
	uint32_t tq;		/* Time quanta (TQ) in nanoseconds */
	uint32_t prop_seg;		/* Propagation segment in TQs */
	uint32_t phase_seg1;	/* Phase buffer segment 1 in TQs */
	uint32_t phase_seg2;	/* Phase buffer segment 2 in TQs */
	uint32_t sjw;		/* Synchronisation jump width in TQs */
	uint32_t brp;		/* Bit-rate prescaler */
};

static int mcp251x_do_set_bittiming(struct mpsse_context* can,
		struct can_bittiming *bt)
{
	write_reg(can, CNF1, ((bt->sjw - 1) << CNF1_SJW_SHIFT) | bt->brp);
	write_reg(can, CNF2, CNF2_BTLMODE |
				((bt->phase_seg1 - 1) << CNF2_PS1_SHIFT) |
				(bt->prop_seg - 1));
	bit_mod_reg(can, CNF3, CNF3_PHSEG2_MASK, (bt->phase_seg2 - 1));
	printf("CNF: 0x%02x 0x%02x 0x%02x\n",
		read_reg(can, CNF1),
		read_reg(can, CNF2),
		read_reg(can, CNF3));
	return 0;
}

static int set_bittiming(struct mpsse_context* can, int bitrate, int osc)
{
/* (start_seg + prop_seg + phase_seg1 + phase_seg2) * tq == 10e9/bitrate
 * const start_seg = 1;
 * must: phase_seg2 >= 2
 */
	struct can_bittiming bt = {0};
	switch(osc){
	case (20*1000*1000):
		switch(bitrate){
		case (1000*1000):
			bt.brp = 1;
			break;
		case (500*1000):
			bt.brp = 3;
			break;
		case (250*1000):
			bt.brp = 7;
			break;
		case (125*1000):
			bt.brp = 15;
			break;
		default:
			fprintf(stderr,
				"Unsupported pair bitrate/osc %i / %i\n",
				bitrate, osc);
			return -EINVAL;
			break;
		}
		break;
	case (25*1000*1000):
		switch(bitrate){
		case (500*1000):
			bt.brp = 4;
			break;
		case (250*1000):
			bt.brp = 9;
			break;
		case (125*1000):
			bt.brp = 19;
			break;
		default:
			fprintf(stderr,
				"Unsupported pair bitrate/osc %i / %i\n",
				bitrate, osc);
			return -EINVAL;
			break;
		}
		break;
	default:
		fprintf(stderr,
			"Unsupported osc value %i\n", osc);
		return -EINVAL;
		break;
	}
	bt.prop_seg = 1;
	bt.phase_seg1 = 1;
	bt.phase_seg2 = 2;
	bt.bitrate = bitrate;
	bt.tq = (1000*1000*1000)/osc * 2*(bt.brp+1);
	bt.sjw = 4;
	printf("Set bitrate %i tq %i brp %i osc %i\n", bitrate, bt.tq,
		bt.brp, osc);
	return mcp251x_do_set_bittiming(can, &bt);
}
// Parses given string: extract filter number and filter elements (sid and eid)
// Returns a found number of filter
static int parse_filter(
	char *description, 	//Filter description given as nn:sid-eid
				//where nn filter/mask number
				// sid and eid in hex
	char *filter)	// points where to store the obtained filter.
			//Must be allocated by caller CAN_ID_SIZE 
{
	uint32_t sid, eid;
	char *point = description;

	unsigned int num = strtol(point, &point, 10);
	if(point[0] != ':') {
		fprintf(stderr, "Malformed filter description %s %s %p %p --\n",
			description, point, description, point);
		return -EINVAL;
	}

	sid = strtol(++point, &point, 16);// In case of error the entry is zeroed
		if(point[0] != '-') {
			fprintf(stderr, "Malformed filter description %s B\n",
				description);
			return -EINVAL;
		}

	eid = strtol(++point, &point, 16);// In case of error the entry is zeroed
		if(point[0] != 0) {
			fprintf(stderr, "Malformed filter description %s C\n",
				description);
			return -EINVAL;
		}
	filter[0] = sid >> 3 & 0xff;//SIDH
	filter[1] = (sid & SIDL_SID_MASK) << SIDL_SID_SHIFT;//SIDL
	if(eid) {
		filter[1] |= 1 << SIDL_EXIDE_SHIFT;//SIDL
		filter[1] |= eid >> SIDL_EXIDE_SHIFT & SIDL_EID_MASK;//SIDL
		filter[2] = eid >> 8 & 0xff; //EID8
		filter[3] = eid & 0xff; //EID0
	}
	else {
		filter[2] = 0; //EID8
		filter[3] = 0; //EID0
	}
	return num; //return the number of the filter described
}
/*
* Probe matching VId:PId devices on USB bus, looks for USB to SPI to CAN bundle
* FT232 -- mcp251x
* Returns a pointer to a mpss context of a first found chain.
*/
static struct mpsse_context* find_mcp251x(int vid, int pid)
{
	int retval = EXIT_FAILURE;
	struct mpsse_context *can = NULL;

	for(unsigned int i=0; i<100 ;++i){
		can = OpenIndex(vid, pid, SPI0, (1000*1000), MSB, IFACE_A,
				NULL, NULL, i);
		if(can != NULL && can->open){
			printf("FTDI chip %i initialized at %dHz (SPI mode 0)\n"
					, i, GetClock(can));
			if(reset_cmd(can)){
				fprintf(stderr,"Reset error\n");
				continue;
			}
			usleep(50000);//Let the slave chip wake after reset

			retval = read_reg(can, CANCTRL);
			if(0 > retval){
				fprintf(stderr,"Read error\n");
				continue;
			}
			else{
				if((retval & CANCTRL_REQOP_MASK) !=
						CANCTRL_REQOP_CONF){
					fprintf(stderr,
						"Device answer doesn't match\n"
						"%x & %x != %x\n",
						retval, CANCTRL_REQOP_MASK,
						CANCTRL_REQOP_CONF);
				}
				else
				{
					printf("Found device ind %i\n", i);
					break;//Time to return
				}
			}
		}
		else
		{// Seemingly the search has finished with no result
			fprintf(stderr, "Failed to initialize MPSSE-%i: %s\n",
					i, ErrorString(can));
		}

		if(can){
			Close(can);
			can = NULL;
		}
	}
	return can;// Real pointer or NULL
}

static int decimal_modifier_2_factor(char modifier)
{
	switch(modifier) {
	case 'k':
	case 'K':
		return 1000;
		break;
	case 'm':
	case 'M':
		return 1000000;
		break;
	default:
		printf ("Unknown modifier `%c'\n", modifier);
	case 0: //endofline, no modifier
		return 1;
		break;
	}
}
void print_received_frame(int rx_buf_num, char* data, int showraw)
{//TODO Make up this mess
	struct timespec tp;
	uint32_t sid, eid, srr, ide, rtr, dlc;
	char buf[255];
	char *str = buf;

	if(!clock_gettime(CLOCK_REALTIME, &tp))
		str += sprintf(str, "%lu.%lu #%i: ",
			tp.tv_sec, tp.tv_nsec, rx_buf_num);
	else
		str += sprintf(str, "#%i: ",rx_buf_num);

	sid = data[0] << 3 | data[1] >> 5;
	str += sprintf(str, "%0x", sid & ((1<<11)-1));
	if(data[1] & 0x08) { //EXIDE
		eid = (data[1] &0x03) << 16 | data[2] << 8 | data[3];
		rtr = data[4] & 0x40;
		srr = 0;
		str += sprintf(str, "-%0x: ", eid & ((1<<19)-1));
	}
	else {
		eid = 0;
		rtr = 0;
		srr = data[1] & 0x10;
		str += sprintf(str, ": ");
	}

	if(rtr || srr)
		str += sprintf(str, "RTR");
	else{
		dlc = data[4] & 0x0f;
		if(dlc > 8)
			str += sprintf(str, "Incorrect DLC");
		else {
			for(int i = 0; i < dlc; ++i)
				str += sprintf(str, "%02x ", data[5 + i] & 0xff);
		}
	}

	if(showraw){
		str += sprintf(str, "== ");
		for(int i = 0; i < 13; ++i)
			str += sprintf(str, "%0x ", data[i] & 0xff);
		str += sprintf(str, "==");
	}

	str += sprintf(str, "\n");

	printf(buf);
	return;
}
#define INPUT_BUFFER_NUM 4 //power of 2
static inline unsigned inc_point(unsigned i, unsigned num)
{return (i+1)&(num-1);}

static char i_buf[INPUT_BUFFER_NUM][CAN_MAX_FRAME];
static unsigned w_point = 0, r_point = 0;

// Keyboard input processing thread
static void *user_input(void *v)
{
//Offsets of specific fields in CAN registers file
#define SIDH 0
#define SIDL 1
#define EID8 2
#define EID0 3
#define DLC 4
#define DAT 5
	char tmp_buffer[1024];
	char *point;
	unsigned sid, eid,dlc;

	for(w_point=0; ;) {
		fgets(tmp_buffer, sizeof(tmp_buffer), stdin);

		if(tmp_buffer[0] == 'q' ||
			tmp_buffer[0] == 'Q')
			exit(0);

		point = tmp_buffer;

		sid = strtol(point, &point, 16);
		i_buf[w_point][SIDH] = sid >> 8 & 0xff;
		i_buf[w_point][SIDL] = (sid & SIDL_SID_MASK) << SIDL_SID_SHIFT;
		if(point[0] == ':') { //Only sid presented
			i_buf[w_point][EID8] = 0;
			i_buf[w_point][EID0] = 0;
		}
		else if(point[0] == '-') { //sid+eid
			eid = strtol(point + 1, &point, 16);
			if(point[0] == ':') { //a correct record termination
				i_buf[w_point][SIDL] |= 1 << SIDL_EXIDE_SHIFT;
				i_buf[w_point][SIDL] |=
					eid >> SIDL_EXIDE_SHIFT & SIDL_EID_MASK;
				i_buf[w_point][EID8] = eid >> 8 & 0xff;
				i_buf[w_point][EID0] = eid & 0xff;
			}
			else {
				goto err;
			}
		}
		else {
			goto err;
		}

		for(dlc=0; dlc <= 8; ++dlc){
			if(point[0])
				i_buf[w_point][dlc + DAT] =
					strtol(point + 1, &point, 16);
			else
				break;
		}

		if(dlc == 8)
			goto err;
		i_buf[w_point][DLC] = dlc - 1;

		while(r_point == inc_point(w_point, INPUT_BUFFER_NUM)){
			fprintf(stderr, "Transmit stall\a\n\a");
			usleep(100000);
		}
		w_point = inc_point(w_point, INPUT_BUFFER_NUM);
		printf("%04x-%04x %04x:", sid, eid, dlc);
		continue;
err:
		fprintf(stderr, "Malformed use input %s char %i\n",
			tmp_buffer, (int)(point-tmp_buffer));
	}
}
// Reads a ready-to-send frame (if any) from input buffer
static char *get_input(int showraw)
{
	int t;
	if(r_point == w_point)
		return NULL;
	else {
		t = r_point;
		r_point = inc_point(r_point, INPUT_BUFFER_NUM);
		if(showraw){
			printf("==");
			for(int i=0; i<16; ++i)
			printf("%02x ",i_buf[t][i] & 0xff);
			printf("==\n");
		}
		return i_buf[t];
	}
}

static struct mpsse_context *can_2be_reset;
static void cleanup(void)
{
	fprintf(stderr, "Stopped by user\n");
	reset_cmd(can_2be_reset);
}
char help_msg[] = {"\n\
Simple CAN sniffer version %s\n\
Prints received frames to STDOUT, sends to net frames gotten from STDIN.\
\n\
Options:\n\
	--can_bitrate Hz	required CAN bus speed\n\
	--can_osc Hz		CAN clock oscillator frequency\n\
\n\
decimal modifiers 'k' and 'M' are allowed\n\
\n\
	--promiscuous [01]	a given RX buffer is to receive all frames\n\
	--rollover		rollover mode for RX buffer 0\n\
	--mask [01]:sid-eid	acceptance mask for a mentioned RX buffer\n\
	--filter [012345]:sid-eid	acceptance filter\n\
\n\
filters 0 and 1 are purposed for RX buffer 0 and the others for RX buffer 1\n\
	mask and filters format: n:sid-eid\n\
	where:\n\
	num	number of mask or filter\n\
	sid	Standard identifier of mask/filter\n\
	eid	enhanced identifier of mask/filter\n\
Omitted eid (if any) will be zeroed.\n\
\n\
	--sniff			listen only mode\n\
	--showraw		print out frames received/to be sent in hex\n\
\n\
Allowed frame to be sent format:\n\
sid:paiload[0...i]\n\
sid-eid:paiload[0...i]\n\
	where i must be less than 8\n\
"};

static void print_help(void)
{
	fprintf(stderr, help_msg, VER);
}
int main(int argc, char **argv)
{
	//struct can_frame frame;
	char *data = NULL;
	unsigned int can_bitrate = (125*1000);
	unsigned int can_osc = (25*1000*1000);
	/* bitmask points which of RX buffers (or both)
	to be configured in promiscuous mode */
	unsigned int promiscuous = 0;
	unsigned int rollover = 0;
	unsigned int showraw = 0;
	uint8_t mode = CANCTRL_REQOP_NORMAL;
	char filter[CAN_FILTER_NUM][CAN_ID_SIZE] = {0};
	char mask[2][CAN_ID_SIZE] = {0};

	while (1){
		char filter_tmp[CAN_ID_SIZE];
		int retval;
		opterr = 1;
		static struct option long_options[] =
		{
			{"can_bitrate",	required_argument, 0, 'b'},
			{"can_osc",	required_argument, 0, 'q'},
			{"promiscuous",	required_argument, 0, 'p'},
			{"rollover",	required_argument, 0, 'r'},
			{"mask",	required_argument, 0, 'm'},
			{"filter",	required_argument, 0, 'f'},
			{"sniff",	no_argument, 0, 's'},
			{"showraw",	no_argument, 0, 'S'},
			{0, 0, 0, 0}
		};
		int option_index = 0;
		int c = getopt_long (argc, argv, "b:q:rm:f:",
				long_options, &option_index);
		if (c == -1)
			break;
		switch (c) {
			char *modifier;
			int val;
		case 'b':
			can_bitrate = strtoul(optarg, &modifier, 10);
			can_bitrate *= decimal_modifier_2_factor(*modifier);
			break;
		case 'q':
			can_osc = strtoul(optarg, &modifier, 10);
			can_osc *= decimal_modifier_2_factor(*modifier);
			break;
		case 'r':
			rollover = 1;
			break;
		case 'p':
			val = atoi(optarg);
			switch(val){
			case 0:
			case 1:
				promiscuous |= 1 << val;
				break;
			default:
				fprintf(stderr,
					"Invalid buffer number %i for promiscuous mode\n",
				val);
				abort();
				break;
			}
			break;
		case 'f':
			val = parse_filter(optarg, filter_tmp);
			switch(val){
			case 0:
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
				memcpy(filter[val], filter_tmp, CAN_ID_SIZE);
				break;
			default:
				fprintf(stderr,
					"Invalid filter number %i\n", val);
				abort();
				break;

			}
			break;
		case 'm':
			val = parse_filter(optarg, filter_tmp);
			switch(val){
			case 0:
			case 1:
				memcpy(mask[val], filter_tmp, CAN_ID_SIZE);
				break;
			default:
				fprintf(stderr,
					"Invalid mask number %i\n", val);
				abort();
				break;

			}
			break;
		case 's':
			mode = CANCTRL_REQOP_LISTEN_ONLY;
			break;
		case 'S':
			showraw = 1;
			break;
		case '?':
			/* getopt_long already printed an error message. */
		default:
			print_help();
			abort();
		}
	}

	struct mpsse_context *can = find_mcp251x(0x0403, 0x6014);

	if(!can){
		fprintf(stderr, "mcp251x not found\n");
		abort();
	}
	can_2be_reset = can;
	atexit(cleanup);

	if(set_bittiming(can, can_bitrate, can_osc))
		abort();

	//Begin the device setup
	bit_mod_reg(can, CANCTRL, CANCTRL_REQOP_MASK, CANCTRL_REQOP_CONF);

	/* Config RX buffers for promiscuous mode and rollover */
	bit_mod_reg(can, RXBCTRL(0), RXBCTRL_BUKT,
		rollover ? RXBCTRL_BUKT : 0);
	for(int i=0; i < 2; ++i){
		bit_mod_reg(can, RXBCTRL(i), RXBCTRL_RXM0 | RXBCTRL_RXM1,
			promiscuous & (1<<i) ?
			RXBCTRL_RXM0 | RXBCTRL_RXM1 : 0);
	}
	for(int i=0; i<CAN_RXBUF_NUM; ++i) {
		write_reg(can, RXMSIDH(i), mask[i][0]);
		write_reg(can, RXMSIDL(i), mask[i][1]);
		write_reg(can, RXMEID8(i), mask[i][2]);
		write_reg(can, RXMEID0(i), mask[i][3]);
		printf("Mask %02x %02x %02x %02x\t",
			mask[i][0] & 0xff, mask[i][1] & 0xff,
			mask[i][2] & 0xff, mask[i][3] & 0xff);
	}
	printf("\n");
	for(int i=0; i<CAN_FILTER_NUM; ++i) {
		write_reg(can, RXFSIDH(i), filter[i][0]);
		write_reg(can, RXFSIDL(i), filter[i][1]);
		write_reg(can, RXFEID8(i), filter[i][2]);
		write_reg(can, RXFEID0(i), filter[i][3]);
		printf("Fltr %02x %02x %02x %02x\t",
			filter[i][0] & 0xff, filter[i][1] & 0xff,
			filter[i][2] & 0xff, filter[i][3] & 0xff);
	}
	printf("\n");

	write_reg(can, CANCTRL, mode);//Go!

	printf("RXBCTRL: 0x%02x 0x%02x CANSTAT 0x%02x\n",
		read_reg(can, RXBCTRL(0)),
		read_reg(can, RXBCTRL(1)),
		read_reg(can, CANSTAT));

	pthread_t input_thread;
	pthread_create(&input_thread, NULL, user_input, NULL);

	while(1){
		int retval, rx_buf_num;
		char *tbuf = get_input(showraw);
		if(tbuf)
			if(!(TXBCTRL_TXREQ & read_reg(can, TXBCTRL(0))))
				send_frame(can, 0, tbuf);

		retval = read_rx_status(can);
		if(0 > retval){
			fprintf(stderr, "Status read error\n");
			break;
		}
		if(retval & RXSTATUS_RXB0)
			rx_buf_num = 0;
		else if(retval & RXSTATUS_RXB1)
			rx_buf_num = 1;
		else
			continue;

		data = read_rx_buffer(can, rx_buf_num);
		if(data){
			print_received_frame(rx_buf_num, data, showraw);
			free(data);
			data = NULL;
		}
	}
}
